;; Generated package description from pinentry.el  -*- no-byte-compile: t -*-
(define-package "pinentry" "0.1.0.20180116.131526" "GnuPG Pinentry server implementation" 'nil :url "https://elpa.gnu.org/packages/pinentry.html" :authors '(("Daiki Ueno" . "ueno@gnu.org")) :maintainer '("Daiki Ueno" . "ueno@gnu.org") :keywords '("gnupg"))
